#!/bin/python3
#%%
import os
import functools
import pandas as pd
import numpy as np
from tools.data_processing.basic_data_tools import multiprocessing_pool_map, pickle_dump, pickle_load
from tools.modeling.evaluation_tools import dataframe_auc_and_ks_score
from tools.modeling.qcut_tools import get_quantile_bins, get_distri_df, get_psi_series

#%%
from tools.std_lib.log import ColorfulLogger
log = ColorfulLogger(basename='feature_tools_v2.py')

#%%


def calculate_gini_impurity(series: pd.Series):
    valid_series = series.dropna()
    if len(valid_series) == 0:
        gini = .0
    else:
        value_counts = valid_series.value_counts()
        probabilities = value_counts / len(valid_series)
        gini = 1 - (probabilities ** 2).sum()
    return gini


def get_trend_correlation(s: pd.Series) -> float:
    """
    Calculate the Spearman correlation between a Series' values and their order to detect trends.

    A positive value indicates an upward trend, a negative value indicates a downward trend,
    and a value near 0 indicates no clear trend.

    Parameters:
    - s (pd.Series): Input Series with numeric values.

    Returns:
    - float: Spearman correlation coefficient between values and their order (range: [-1, 1]).

    Raises:
    - TypeError: If s is not a pandas Series.
    - ValueError: If s is empty or contains no valid numeric data.
    """
    # 输入验证
    if not isinstance(s, pd.Series):
        raise TypeError(f"Input must be a pandas Series but s = {s}")
    # if len(s) == 0:
    #     raise ValueError("Series cannot be empty")
    # if s.dropna().empty:
    #     raise ValueError("Series contains no valid (non-NaN) numeric data")

    corr = float('nan')

    if len(s) <= 1:
        corr = float('nan')
    else:
        s = s.dropna()
        if s.nunique() == 1:
            corr = 0
        else:
            corr = s.corr(pd.Series(range(1, len(s) + 1), index=s.index), method='spearman')

    return corr


def get_quantiles_dict(
        data_df: pd.DataFrame, 
        feature_names: list, 
        n_interval = 10,
        disable_multiprocessing = False,
        ):
    func_name = 'get_quantiles_dict()'
    log.info('started', func = func_name)
    quantiles_dict = multiprocessing_pool_map(
        # get_quantile_bins_func, 
        functools.partial(get_quantile_bins, n = n_interval, drop_endpoints=True, add_endpoints='inf'),
        data_df[feature_names],
        disable_multiprocessing=disable_multiprocessing
        )
    log.info('completed', func = func_name)
    quantiles_dict = pd.Series(quantiles_dict)
    return quantiles_dict


def get_grouped_distri_long_df(
        data_df: pd.DataFrame, 
        quantiles_dict: pd.Series ,
        group_cols: list = None, 
        # feature_names: pd.Series, 
        disable_multiprocessing = False,
        ):
    feature_names = quantiles_dict.index
    func_name = 'get_grouped_distri_long_df()'
    log.info(f'started, group_cols = {group_cols}', func = func_name)
    if group_cols is None:
        stat_df = get_distri_dict(
            data_df=data_df, 
            quantiles_dict=quantiles_dict,
            disable_multiprocessing=disable_multiprocessing,
            )
        stat_df.reset_index(inplace=True)
        stat_df.set_index(['feature_name', 'interval'], inplace = True)
        stat_df.columns.name = 'stat'
    else:
        gp = data_df.groupby(by = group_cols)
        stat_df = gp[feature_names].apply(lambda df: get_distri_dict(
            data_df=df, 
            quantiles_dict=quantiles_dict,
            disable_multiprocessing=disable_multiprocessing,
            ))
        stat_df.reset_index(inplace=True)
        stat_df.set_index(['feature_name'] + group_cols + ['interval'], inplace = True)
        stat_df.columns.name = 'stat'
    log.info(f'completed, group_cols = {group_cols}', func = func_name)
    return stat_df


def get_distri_df_func(series_and_quantiles: tuple):
    series, quantiles = series_and_quantiles
    distri_df = get_distri_df(series, quantiles, NaN=True)
    return distri_df


def get_distri_dict(
        data_df: pd.DataFrame, 
        quantiles_dict: pd.Series, 
        disable_multiprocessing = False
        ):
    vars = {
        col: (data_df[col], quantiles_dict[col]) 
        for col in quantiles_dict.index
        }
    distri_dict = multiprocessing_pool_map(
        get_distri_df_func, 
        vars, 
        disable_multiprocessing=disable_multiprocessing
        )
    distri_dict = pd.concat(distri_dict)
    distri_dict.sort_index(inplace=True)
    distri_dict.index.names = ['feature_name', 'interval']
    # print()
    return distri_dict


###################################################################



def flatten_header_3to2(row: pd.Series):
    return pd.Series({
        'stat': row['stat'] + '-' + str(row['y']),
        'month': row['month']
    })


def concat_wide_dfs(wide1_df: pd.DataFrame, wide2_df: pd.DataFrame):
    if wide1_df.index.equals(wide2_df.index) and wide1_df.columns.intersection(wide2_df.columns).empty and wide1_df.columns.names == wide2_df.columns.names:
        distri_long_df = pd.concat([wide1_df, wide2_df], axis=1)
        if distri_long_df.isna().any().any():
            raise ValueError("There are NaN values in the combined DataFrame")
    else:
        raise ValueError("Indexes of wide1_df and wide2_df do not match")
    return distri_long_df


def get_sum_feature_wide_df(feature_wide_df: pd.DataFrame):
    sum_feature_wide_df = pd.DataFrame({
        col: [feature_wide_df[col].sum()]
        for col in feature_wide_df.columns
    }, index=['total'])
    sum_feature_wide_df.columns.names = ['stat', 'month']
    sum_feature_wide_df.index.name = 'interval'
    return sum_feature_wide_df


def get_sum_wide_df(distri_long_df: pd.DataFrame):
    sum_wide_df = distri_long_df.groupby(level='feature_name').apply(get_sum_feature_wide_df)
    for key in ['percent', 'percent-0', 'percent-1']:
        if key not in sum_wide_df.columns.get_level_values(level='stat'):
            # log.warning(f'{key} not in sum_wide_df, skipping normalization for this stat. Please check the data.', func = 'get_sum_wide_df')
            continue
        tmp_df = sum_wide_df.xs(level = 'stat', key = key, axis = 1)
        if (tmp_df - 1).abs().max().max() < 10e-12:
            cols = sum_wide_df.columns.get_level_values(level='stat') == key
            sum_wide_df.loc[:, cols] = 1
        else:
            raise ValueError('Percent columns should be close to 1.0')
    return sum_wide_df


def sort_tag(var: object):
    if isinstance(var, pd.Interval):
        result = 0
    elif isinstance(var, str):
        if var == 'NaN':
            result = 1
        elif var == 'total':
            result = 2
        else:
            raise ValueError(f"Unexpected value: {var}")
    return result



def get_sort_multiindex(index: pd.Index):
    df = index.to_frame().reset_index(drop=True)
    df['sort_tag'] = df.apply(lambda row: sort_tag(row['interval']), axis=1)
    df.sort_values(by=['feature_name', 'sort_tag', 'interval'], inplace=True)
    df.set_index(['feature_name', 'interval'], inplace=True)
    sort_index = df.index
    return sort_index


def multiindex_loc_insert(
        data_df: pd.DataFrame, 
        loc: int, 
        column: str,
        value: object):
    # if not isinstance(data_df.index, pd.MultiIndex):
    #     raise ValueError("data_df must have a MultiIndex")
    index_df = data_df.index.to_frame(index=False).copy()
    index_df.insert(loc, column, value)
    index_df.set_index(keys = index_df.columns.to_list(), inplace=True)
    return index_df.index


def get_no_y_distri_long_df(
        data_df: pd.DataFrame,
        # feature_names: list,
        quantile_dict: pd.Series,
        disable_multiprocessing = False,
    ):
    # feature_names = quantile_dict.index
    func_name = 'get_no_y_distri_long_df()'
    log.info('started', func = func_name)

    group_cols = ['month']
    month_distri_long_df = get_grouped_distri_long_df(
        data_df=data_df,
        quantiles_dict=quantile_dict,
        group_cols=group_cols,
        # feature_names=feature_names,
        disable_multiprocessing=disable_multiprocessing
    )
    no_month_distri_long_df = get_grouped_distri_long_df(
        data_df=data_df,
        quantiles_dict=quantile_dict,
        # group_cols=group_cols,
        # feature_names=feature_names,
        disable_multiprocessing=disable_multiprocessing
    )
    no_month_distri_long_df.index = multiindex_loc_insert(no_month_distri_long_df, 1, 'month', 'total')
    
    month_distri_long_df = month_distri_long_df.unstack(level='month', fill_value=0)
    no_month_distri_long_df = no_month_distri_long_df.unstack(level='month', fill_value=0)
    # print(no_month_distri_long_df)

    # no_total_distri_long_df = month_distri_long_df
    no_total_distri_long_df = pd.concat([
        no_month_distri_long_df, 
        month_distri_long_df
        ], axis = 1)
    no_total_distri_long_df.sort_index(axis=1, inplace=True)

    sum_wide_df = get_sum_wide_df(no_total_distri_long_df)
    distri_long_df = pd.concat([no_total_distri_long_df, sum_wide_df], axis=0)

    sort_multiindex = get_sort_multiindex(distri_long_df.index)
    distri_long_df = distri_long_df.loc[sort_multiindex]

    log.info('completed', func = func_name)

    return distri_long_df


def get_y_distri_long_df(
        data_df: pd.DataFrame,
        quantile_dict: pd.Series,
        # feature_names: list,
        disable_multiprocessing = False,
    ):
    feature_names = quantile_dict.index
    func_name = 'get_y_distri_long_df()'
    log.info('started', func = func_name)

    if data_df['y'].nunique(dropna=True) == 2:

        group_cols = [ 'y', 'month', ]
        y_month_distri_long_df = get_grouped_distri_long_df(
            data_df=data_df,
            quantiles_dict=quantile_dict,
            group_cols=group_cols,
            # feature_names=feature_names,
            disable_multiprocessing=disable_multiprocessing
        )
        no_month_distri_long_df = get_grouped_distri_long_df(
            data_df=data_df,
            quantiles_dict=quantile_dict,
            group_cols=['y'],
            # feature_names=feature_names,
            disable_multiprocessing=disable_multiprocessing
        )
        no_month_distri_long_df.index = multiindex_loc_insert(no_month_distri_long_df, 2, 'month', 'total')
        
        y_month_distri_long_df = y_month_distri_long_df.unstack(level=['y', 'month',], fill_value=0)
        y_month_distri_long_df.columns = y_month_distri_long_df.columns.to_frame().apply(flatten_header_3to2, axis=1).set_index(['stat', 'month']).index
        
        no_month_distri_long_df = no_month_distri_long_df.unstack(level=['y', 'month',], fill_value=0)
        no_month_distri_long_df.columns = no_month_distri_long_df.columns.to_frame().apply(flatten_header_3to2, axis=1).set_index(['stat', 'month']).index
        # print(no_month_distri_long_df)

        # no_total_distri_long_df = y_month_distri_long_df
        no_total_distri_long_df = pd.concat([
            no_month_distri_long_df, 
            y_month_distri_long_df
            ], axis = 1)
        no_total_distri_long_df.sort_index(axis=1, inplace=True)

        sum_wide_df = get_sum_wide_df(no_total_distri_long_df)
        distri_long_df = pd.concat([no_total_distri_long_df, sum_wide_df], axis=0)

        sort_multiindex = get_sort_multiindex(distri_long_df.index)
        distri_long_df = distri_long_df.loc[sort_multiindex]

    else:
        log.error('Error in get_y_month_distri_long_df: y must be binary', func = func_name)
        raise ValueError('y must be binary')

    log.info('completed', func = func_name)

    return distri_long_df


def get_monthly_psi_from_feature_distri_long_df(feature_distri_long_df: pd.DataFrame):

    percent_df = feature_distri_long_df.xs(level='stat', key = 'percent', axis = 1).copy()
    percent_df = percent_df.loc[percent_df.index != 'total', :]

    psi_df = {}
    months = sorted(percent_df.columns.to_list())
    if len(months) > 1:
        for a_month in months:
            for b_month in months:
                if a_month >= b_month:
                    continue
                tmp_df = get_psi_series(
                    percent_df[a_month],
                    percent_df[b_month],
                    return_type='series',
                )
                psi_df[f'{a_month}-vs-{b_month}'] = tmp_df
        psi_df = pd.concat(psi_df, axis = 1)
        psi_df.loc['total'] = psi_df.sum(axis=0)

        psi_df.columns.name = 'month'
        cols_df = psi_df.columns.to_frame()
        cols_df.insert(0, 'stat', 'psi')
        cols_df.set_index(['stat', 'month'], inplace=True)
        psi_df.columns = cols_df.index

    else:
        # log.warning('len(months) == 1')
        psi_df = pd.DataFrame(
            index=feature_distri_long_df.index, 
            columns = pd.MultiIndex.from_tuples(
                [], names=feature_distri_long_df.columns.names
            ))
        
    return psi_df


def get_monthly_iv_from_feature_distri_long_df(feature_distri_long_df: pd.DataFrame):

    months = feature_distri_long_df.xs(level='stat', key='count', axis=1).columns

    iv_df = {}
    for month in months:
        percent_df = feature_distri_long_df.xs(level='month', key=month, axis=1)[['percent-0', 'percent-1']].copy()
        percent_df = percent_df.loc[percent_df.index != 'total', :]
        iv_df[month] = get_psi_series(percent_df['percent-0'], percent_df['percent-1'], return_type='series')
    iv_df = pd.DataFrame(iv_df)
    iv_df.loc['total'] = iv_df.sum(axis=0)

    iv_df.columns.name = 'month'
    cols_df = iv_df.columns.to_frame()
    cols_df.insert(0, 'stat', 'iv')
    cols_df.set_index(['stat', 'month'], inplace=True)
    iv_df.columns = cols_df.index

    return iv_df


def get_expanded_header(df_columns: pd.DataFrame, name: str) -> pd.MultiIndex:
    if isinstance(df_columns, pd.DataFrame):
        df_columns = df_columns.columns
    cols_df = df_columns.to_frame(name='month')
    cols_df.insert(0, 'stat', name)
    cols_df.set_index(['stat', 'month'], inplace=True)
    return cols_df.index


def get_perf_from_feature_distri_long_df(feature_distri_long_df: pd.DataFrame):

    pos_pct_df = feature_distri_long_df.xs(level='stat', key='count-1', axis=1) / feature_distri_long_df.xs(level='stat', key='count', axis=1)
    lift_df = pos_pct_df / pos_pct_df.loc['total']
    pos_pct_df.columns = get_expanded_header(pos_pct_df, 'pos_pct')

    lift_df.columns = get_expanded_header(lift_df, 'lift')

    woe_df = (
        feature_distri_long_df.xs(level='stat', key='percent-1', axis=1) / feature_distri_long_df.xs(level='stat', key='percent-0', axis=1).replace(0.0, 1e-6)
        ).apply(np.log)
    woe_df.columns = get_expanded_header(woe_df, 'woe')

    perf_df = pd.concat([pos_pct_df, lift_df, woe_df], axis = 1)

    return perf_df


def get_feature_stats_wide_df(feature_distri_long_df: pd.DataFrame):
    psi_df = get_monthly_psi_from_feature_distri_long_df(feature_distri_long_df)
    if 'count-0' in feature_distri_long_df.columns and 'count-1' in feature_distri_long_df.columns:
        iv_df = get_monthly_iv_from_feature_distri_long_df(feature_distri_long_df)
        perf_df = get_perf_from_feature_distri_long_df(feature_distri_long_df)
        feature_stats_df = pd.concat([psi_df, iv_df, perf_df], axis = 1)
    else:
        feature_stats_df = psi_df

    return feature_stats_df


def get_features_stats_long_df(
    distri_long_df: pd.DataFrame,
    disable_multiprocessing = False
    ):
    vars = {
        key: distri_long_df.xs(level='feature_name', key=key)
        for key in distri_long_df.index.get_level_values('feature_name').unique()
    }
    features_stats_long_df = multiprocessing_pool_map(
        func=get_feature_stats_wide_df,
        iterable=vars,
        disable_multiprocessing=disable_multiprocessing, 
    )
    features_stats_long_df = pd.concat(features_stats_long_df)
    features_stats_long_df.index.names = ['feature_name', 'interval']
    return features_stats_long_df


def get_months_from_dataframe(data_df: pd.DataFrame, pattern = r'\d{4}-\d{2}'):
    return data_df.loc[[], data_df.columns.str.contains(pattern)].columns


def stat_monthly_wide_df(total_features_stats_long_df, stat_name):

    if isinstance(total_features_stats_long_df.columns, pd.MultiIndex):
        if stat_name in total_features_stats_long_df.columns.get_level_values('stat'):
            origin_total_and_monthly_df = total_features_stats_long_df.xs(level = 'stat', key = stat_name, axis = 1).copy()
        else:
            origin_total_and_monthly_df = pd.DataFrame(
                index=total_features_stats_long_df.index, 
                columns=pd.Index([], name='month')
                )
    else:
        origin_total_and_monthly_df = total_features_stats_long_df.copy()

    # get_trend_correlation_no_total = lambda s: get_trend_correlation(s.loc[s.index != 'total'])
    # get_trend_correlation_no_total = lambda s: get_trend_correlation(s.loc[s.index.str.contains(r'\d{4}-\d{2}')]) if len(s) > 0 else float('nan')
    # month_cols = origin_total_and_monthly_df.columns.str.contains(r'\d{4}-\d{2}')
    month_cols = get_months_from_dataframe(origin_total_and_monthly_df)
    origin_monthly_df = origin_total_and_monthly_df.loc[:, month_cols]

    origin_monthly_df.columns.name = 'stat'
    stat_monthly_df = pd.DataFrame({
        'min': origin_monthly_df.min(axis = 1),
        'mean': origin_monthly_df.mean(axis = 1),
        'max': origin_monthly_df.max(axis = 1),
        'std': origin_monthly_df.std(axis = 1),
        'trend': origin_monthly_df.apply(
            get_trend_correlation, axis = 1 
            ) if 'psi' not in stat_name else float('nan')
    })
    stat_monthly_df.columns.name = 'stat'
    stat_monthly_df['abs_gap'] = stat_monthly_df['max'] - stat_monthly_df['min']
    stat_monthly_df['rel_gap'] = stat_monthly_df['abs_gap'] / stat_monthly_df['mean']
    stat_monthly_df['cv'] = stat_monthly_df['std'] / stat_monthly_df['mean']
    monthly_df = pd.concat([origin_total_and_monthly_df, stat_monthly_df], axis = 1)
    monthly_df.columns = get_expanded_header(monthly_df, stat_name)

    return monthly_df


def get_auc_ks_df(
        data_df: pd.DataFrame, 
        feature_names: list, 
        disable_multiprocessing = False
        ):
    gp = data_df.groupby(by = 'month')
    stat_df = gp.apply(lambda df: dataframe_auc_and_ks_score(
        data_df=df,
        x_cols=feature_names,
        disable_multiprocessing=disable_multiprocessing
    ))
    no_month_stat_df = dataframe_auc_and_ks_score(
        data_df=data_df,
        x_cols=feature_names,
        disable_multiprocessing=disable_multiprocessing
    )
    no_month_stat_df.index = multiindex_loc_insert(no_month_stat_df, 0, 'month', 'total')
    stat_df = pd.concat([no_month_stat_df, stat_df], axis=0)
    stat_df = stat_df.unstack(level='month')
    stat_df.columns.names = ['stat', 'month']
    return stat_df


def get_coverage_df(data_df: pd.DataFrame, feature_names: list):
    func_name = 'get_coverage_df()'
    log.info('started', func = func_name)
    coverage_df = data_df.groupby('month').apply(lambda df: df[feature_names].notna().mean()).T
    total_coverage_df = data_df[feature_names].notna().mean().to_frame('total')
    coverage_df = pd.concat([coverage_df, total_coverage_df], axis=1)
    coverage_df.columns = get_expanded_header(coverage_df, name='coverage')
    log.info('completed', func = func_name)
    return coverage_df


def get_gini_impurity_df(data_df: pd.DataFrame, feature_names: list):
    func_name = 'get_gini_impurity_df()'
    log.info('started', func = func_name)
    gini_impurity_df = data_df.groupby('month').apply(lambda df: df[feature_names].apply(calculate_gini_impurity)).T
    total_gini_impurity_df = data_df[feature_names].notna().mean().to_frame('total')
    gini_impurity_df = pd.concat([gini_impurity_df, total_gini_impurity_df], axis=1)
    gini_impurity_df.columns = get_expanded_header(gini_impurity_df, name='gini_impurity')
    log.info('completed', func = func_name)
    return gini_impurity_df


def get_distri_long_df(
        data_df: pd.DataFrame, 
        quantile_dict: pd.Series,  
        disable_multiprocessing = False
        ):

    no_y_distri_long_df = get_no_y_distri_long_df(
        data_df=data_df, 
        quantile_dict=quantile_dict,
        disable_multiprocessing=disable_multiprocessing
        )
    if data_df['y'].nunique(dropna=True) == 2:
        y_distri_long_df = get_y_distri_long_df(
            data_df=data_df, 
            quantile_dict=quantile_dict,
            disable_multiprocessing=disable_multiprocessing
            )
    else:
        y_distri_long_df = pd.DataFrame(
            index=no_y_distri_long_df.index,
            columns=pd.MultiIndex.from_tuples(
                [], names=no_y_distri_long_df.columns.names)
        )
    distri_long_df = pd.concat(
        [no_y_distri_long_df, y_distri_long_df],
        axis=1
        )
    return distri_long_df



###################################################################


def check_and_fix_feature_data_df(data_df: pd.DataFrame):
    func_name = 'check_and_fix_feature_data_df()'
    if data_df.index.name != 'index':
        raise ValueError()
    if data_df.columns.name != 'feature_name':
        raise ValueError()
    for col in ['date', 'month', 'data_split']:
        if col not in data_df.columns:
            raise ValueError()
    if 'y' in data_df.columns:
        n1 = len(data_df)
        data_df = data_df.loc[data_df['y'].notna(), :]
        n2 = len(data_df)
        if n1 - n2 > 0:
            log.warning(f'found NaN in y col and fixed. ({n1} - {n1 - n2} = {n2})', func = func_name)
        if data_df['y'].nunique(dropna=True) != 2:
            error_str = f'y column must have exactly 2 unique values. Found {data_df["y"].nunique(dropna=True)} unique values.'
            log.error(error_str)
            raise ValueError(error_str)
    else:
        log.warning("No y column in the dataframe. Adding a dummy y column.")
        data_df['y'] = float('nan')
    return data_df


def get_feature_interval_trend(feature_woe_df: pd.DataFrame):
    if not feature_woe_df.index.is_monotonic_increasing:
        log.error('not feature_woe_df.index.is_monotonic_increasing')
        raise ValueError()
    return feature_woe_df.apply(get_trend_correlation)


def get_interval_trend_from_woe_long_df(
        woe_long_df: pd.DataFrame, 
        disable_multiprocessing = False
        ):
    func_name = 'get_interval_trend_from_woe_long_df()'
    log.info(f'started', func = func_name)
    level_index = woe_long_df.index.get_level_values(level = 'interval')
    drop_tags = sorted(set([p for p in level_index if not isinstance(p, pd.Interval)]))
    level_index = ~ level_index.isin(drop_tags)
    woe_long_df = woe_long_df.loc[level_index, :]

    # gp = woe_long_df.groupby(level='feature_name')
    # interval_trend_df = gp.apply(get_feature_interval_trend)
    vars = {
        feature: woe_long_df.loc[feature] 
        for feature in sorted(woe_long_df.index.get_level_values('feature_name').unique())
        }
    interval_trend_df = multiprocessing_pool_map(
        get_feature_interval_trend, 
        vars, 
        to_frame = True,
        disable_multiprocessing=disable_multiprocessing
        ).T
    interval_trend_df.index.name = 'feature_name'
    log.info(f'completed', func = func_name)
    return interval_trend_df


####################################################################






class MonthlyFeatureStats(object):

    def __init__(self, 
                 data_df, 
                 feature_names, 
                 path = None,
                 use_cache = True,
                 n_interval = 10,
                 disable_multiprocessing = False, 
                 ):
        func_name = 'MonthlyFeatureStats.init()'
        data_df = check_and_fix_feature_data_df(data_df)
        self.data_df = data_df
        self.feature_names = feature_names
        self.path = path
        self.n_interval = n_interval
        if (path is not None) and os.path.isfile(path=path) and use_cache:
            log.info('path isfile', func = func_name)
            self.stat = pickle_load(path)
        else:
            self.stat = {}
        self.disable_multiprocessing = disable_multiprocessing
        self.n_months = data_df['month'].nunique(dropna=True)

        if self.n_months < 2:
            log.warning('n_months < 2', func = func_name)

        return None
    

    def dump(self, makedirs = False):
        path = self.path
        stat = self.stat
        if makedirs:
            os.makedirs(os.path.dirname(path), exist_ok=True)
        pickle_dump(stat, path)
    
    def get_quantiles_dict(self):
        if 'quantile_dict' not in self.stat:
            data_df = self.data_df
            feature_names = self.feature_names
            n_interval = self.n_interval
            disable_multiprocessing = self.disable_multiprocessing
            quantile_dict = get_quantiles_dict(
                data_df=data_df,
                feature_names=feature_names,
                n_interval=n_interval,
                disable_multiprocessing=disable_multiprocessing
            )
            self.stat['quantile_dict'] = quantile_dict
        quantile_dict = self.stat['quantile_dict']
        return quantile_dict
    
    def get_distri_long_df(self):
        if 'distri_long_df' not in self.stat:
            data_df = self.data_df
            quantile_dict = self.get_quantiles_dict()
            disable_multiprocessing = self.disable_multiprocessing
            distri_long_df = get_distri_long_df(
                data_df=data_df, 
                quantile_dict=quantile_dict, 
                disable_multiprocessing=disable_multiprocessing
                )
            self.stat['distri_long_df'] = distri_long_df
        distri_long_df = self.stat['distri_long_df']
        return distri_long_df

    def get_features_stats_long_df(self):
        if 'features_stats_long_df' not in self.stat:
            distri_long_df = self.get_distri_long_df()
            disable_multiprocessing = self.disable_multiprocessing
            features_stats_long_df = get_features_stats_long_df(
                distri_long_df,
                disable_multiprocessing=disable_multiprocessing
                )
            self.stat['features_stats_long_df'] = features_stats_long_df
        features_stats_long_df = self.stat['features_stats_long_df']
        return features_stats_long_df

    def get_total_features_stats_wide_df(self, dropna = False):
        func_name = 'get_total_features_stats_wide_df()'
        log.info('Starting to get total features stats wide df.', func=func_name)
        if 'total_features_stats_wide_df' not in self.stat:
            log.info('Calculating total features stats wide df.', func=func_name)
            features_stats_long_df = self.get_features_stats_long_df()
            total_features_stats_wide_df = features_stats_long_df.xs(
                level = 'interval', key = 'total')
            self.stat['total_features_stats_wide_df'] = total_features_stats_wide_df
        total_features_stats_wide_df = self.stat['total_features_stats_wide_df']
        log.info('Finished getting total features stats wide df.', func=func_name)
        if dropna:
            total_features_stats_wide_df = total_features_stats_wide_df.dropna(how='all', axis=1,)
        return total_features_stats_wide_df

    def get_coverage_df(self):
        if 'coverage_df' not in self.stat:
            data_df = self.data_df
            feature_names = self.feature_names
            coverage_df = get_coverage_df(
                data_df, 
                feature_names=feature_names)
            self.stat['coverage_df'] = coverage_df
        coverage_df = self.stat['coverage_df']
        return coverage_df

    def get_gini_impurity_df(self):
        if 'gini_impurity_df' not in self.stat:
            data_df = self.data_df
            feature_names = self.feature_names
            gini_impurity_df = get_gini_impurity_df(
                data_df, 
                feature_names=feature_names)
            self.stat['gini_impurity_df'] = gini_impurity_df
        gini_impurity_df = self.stat['gini_impurity_df']
        return gini_impurity_df

    def get_auc_ks_df(self):
        if 'auc_ks_df' not in self.stat:
            data_df = self.data_df
            feature_names = self.feature_names
            disable_multiprocessing = self.disable_multiprocessing
            auc_ks_df = get_auc_ks_df(
                data_df, 
                feature_names=feature_names, 
                disable_multiprocessing = disable_multiprocessing)
            self.stat['auc_ks_df'] = auc_ks_df
        auc_ks_df = self.stat['auc_ks_df']
        return auc_ks_df

    def get_coverage_wide_df(self):
        if 'coverage_wide_df' not in self.stat:
            coverage_df = self.get_coverage_df()
            coverage_wide_df = stat_monthly_wide_df(coverage_df, 'coverage')
            self.stat['coverage_wide_df'] = coverage_wide_df
        coverage_wide_df = self.stat['coverage_wide_df']
        return coverage_wide_df

    def get_gini_impurity_wide_df(self):
        if 'gini_impurity_wide_df' not in self.stat:
            gini_impurity_wide_df = self.get_gini_impurity_df()
            gini_impurity_wide_df = stat_monthly_wide_df(gini_impurity_wide_df, 'gini_impurity')
            self.stat['gini_impurity_wide_df'] = gini_impurity_wide_df
        gini_impurity_wide_df = self.stat['gini_impurity_wide_df']
        return gini_impurity_wide_df
    
    def get_auc_wide_df(self):
        func_name = 'get_auc_wide_df()'
        if 'auc_wide_df' not in self.stat:
            log.info('started', func = func_name)
            auc_ks_df = self.get_auc_ks_df()
            auc_wide_df = stat_monthly_wide_df(auc_ks_df, 'auc')
            self.stat['auc_wide_df'] = auc_wide_df
            log.info('completed', func = func_name)
        auc_wide_df = self.stat['auc_wide_df']
        return auc_wide_df
    
    def get_ks_wide_df(self):
        func_name = 'get_ks_wide_df()'
        if 'ks_wide_df' not in self.stat:
            log.info('started', func = func_name)
            ks_wide_df = self.get_auc_ks_df()
            ks_wide_df = stat_monthly_wide_df(ks_wide_df, 'ks')
            self.stat['ks_wide_df'] = ks_wide_df
            log.info('completed', func = func_name)
        ks_wide_df = self.stat['ks_wide_df']
        return ks_wide_df
    
    def get_psi_wide_df(self):
        if 'psi_wide_df' not in self.stat:
            total_features_stats_wide_df = self.get_total_features_stats_wide_df()
            psi_wide_df = stat_monthly_wide_df(total_features_stats_wide_df, 'psi')
            self.stat['psi_wide_df'] = psi_wide_df
        psi_wide_df = self.stat['psi_wide_df']
        return psi_wide_df


    def get_adjacent_psi_wide_df(self):

        if 'adjacent_psi_wide_df' not in self.stat:

            total_features_stats_wide_df = self.get_total_features_stats_wide_df()
            psi_df = total_features_stats_wide_df.xs(level = 'stat', key = 'psi', axis=1)
            months = sorted(set(psi_df.columns.str.split('-vs-').to_series().sum()))

            adjacent_months = [f'{m0}-vs-{m1}' for m0, m1 in zip(months[:-1], months[1:])]
            # first_months = [f'{months[0]}-vs-{m1}' for m1 in months[1:]]

            stat_name = 'adjacent_psi'
            adjacent_psi_df = psi_df[adjacent_months].copy()
            adjacent_psi_df.columns = get_expanded_header(adjacent_psi_df.columns, 'adjacent_psi')
            adjacent_psi_wide_df = stat_monthly_wide_df(adjacent_psi_df, stat_name)
            self.stat['adjacent_psi_wide_df'] = adjacent_psi_wide_df
        
        adjacent_psi_wide_df = self.stat['adjacent_psi_wide_df']

        return adjacent_psi_wide_df


    def get_first_psi_wide_df(self):

        if 'first_psi_wide_df' not in self.stat:

            total_features_stats_wide_df = self.get_total_features_stats_wide_df()
            psi_df = total_features_stats_wide_df.xs(level = 'stat', key = 'psi', axis=1)
            months = sorted(set(psi_df.columns.str.split('-vs-').to_series().sum()))
            # adjacent_months = [f'{m0}-vs-{m1}' for m0, m1 in zip(months[:-1], months[1:])]
            first_months = [f'{months[0]}-vs-{m1}' for m1 in months[1:]]
            stat_name = 'first_psi'
            first_psi_df = psi_df[first_months].copy()
            first_psi_df.columns = get_expanded_header(first_psi_df.columns, 'first_psi')
            first_psi_wide_df = stat_monthly_wide_df(first_psi_df, stat_name)
            self.stat['first_psi_wide_df'] = first_psi_wide_df
        
        first_psi_wide_df = self.stat['first_psi_wide_df']

        return first_psi_wide_df

    def get_iv_wide_df(self):
        if 'iv_wide_df' not in self.stat:
            total_features_stats_wide_df = self.stat['total_features_stats_wide_df']
            stat_name = 'iv'
            iv_wide_df = stat_monthly_wide_df(total_features_stats_wide_df, stat_name)
            self.stat['iv_wide_df'] = iv_wide_df
        iv_wide_df = self.stat['iv_wide_df']
        return iv_wide_df
    
    def get_corrected_auc_wide_df(self):
        func_name = 'get_corrected_auc_wide_df()'
        if 'corrected_auc_wide_df' not in self.stat:
            log.info('started', func = func_name)
            auc_ks_df = self.get_auc_ks_df()
            auc_df = auc_ks_df['auc'].copy()

            corr_dir = pd.Series(data = 'unknown', index = auc_df.index)
            corr_dir[auc_df['total'] > .5] = 'maybe_same'
            corr_dir[auc_df['total'] < .5] = 'maybe_opposite'
            corr_dir[auc_df.min(axis = 1) > .5] = 'same'
            corr_dir[auc_df.max(axis = 1) < .5] = 'opposite'

            rows = []
            for _, row in auc_df.iterrows():
                if corr_dir[row.name] in ['same', 'maybe_same', 'unknown']:
                    rows.append(row)
                elif corr_dir[row.name] in ['opposite', 'maybe_opposite']:
                    rows.append(1 - row)
                else:
                    raise ValueError('')

            corrected_auc_df = pd.concat(rows, axis = 1).T
            corrected_auc_df.index.name = 'feature_name'
            corrected_auc_wide_df = stat_monthly_wide_df(corrected_auc_df, stat_name = 'corrected_auc')
            corrected_auc_wide_df[('corrected_auc', 'corr_dir')] = corr_dir
            self.stat['corrected_auc_wide_df'] = corrected_auc_wide_df
            log.info('completed', func = func_name)

        corrected_auc_wide_df = self.stat['corrected_auc_wide_df']

        return corrected_auc_wide_df
    

    def get_interval_trend_wide_df(self):
        func_name = 'get_interval_trend_wide_df()'
        if 'interval_trend_wide_df' not in self.stat:
            log.info(f'started', func = func_name)
            features_stats_long_df = self.get_features_stats_long_df()
            woe_long_df = features_stats_long_df['woe'].copy()
            interval_trend_df = get_interval_trend_from_woe_long_df(
                woe_long_df, 
                disable_multiprocessing=self.disable_multiprocessing
                )
            interval_trend_wide_df = stat_monthly_wide_df(interval_trend_df, 'interval_trend')
            self.stat['interval_trend_wide_df'] = interval_trend_wide_df
            log.info(f'completed', func = func_name)
        interval_trend_wide_df = self.stat['interval_trend_wide_df']
        return interval_trend_wide_df


    def get_abs_interval_trend_wide_df(self):
        func_name = 'get_abs_interval_trend_wide_df()'
        if 'abs_interval_trend_wide_df' not in self.stat:
            log.info(f'started', func = func_name)
            features_stats_long_df = self.get_features_stats_long_df()
            woe_long_df = features_stats_long_df['woe'].copy()
            interval_trend_df = get_interval_trend_from_woe_long_df(
                woe_long_df, 
                disable_multiprocessing=self.disable_multiprocessing
                )
            corr_dir = pd.Series('unknown', interval_trend_df.index)
            corr_dir.loc[interval_trend_df['total'] <= 0] = 'maybe_opposite'
            corr_dir.loc[interval_trend_df['total'] >= 0] = 'maybe_same'
            corr_dir.loc[interval_trend_df.max(axis = 1) <= 0] = 'opposite'
            corr_dir.loc[interval_trend_df.min(axis = 1) >= 0] = 'same'

            abs_interval_trend_df = interval_trend_df.abs()
            abs_interval_trend_wide_df = stat_monthly_wide_df(
                abs_interval_trend_df, 
                'abs_interval_trend'
                )
            abs_interval_trend_wide_df[('abs_interval_trend', 'corr_dir')] = corr_dir
            self.stat['abs_interval_trend_wide_df'] = abs_interval_trend_wide_df
            log.info(f'completed', func = func_name)
        abs_interval_trend_wide_df = self.stat['abs_interval_trend_wide_df']
        return abs_interval_trend_wide_df


    def get_combined_feature_stats_wide_df(self,
                                           feature_importance_df: pd.DataFrame = None,
                                           dropna = False
                                           ):
        func_name = 'get_combined_feature_stats_wide_df()'
        log.info("Combining feature stats into a wide DataFrame.", func = func_name)
        total_features_stats_wide_df = self.get_total_features_stats_wide_df()
        total_features_stats_wide_df = total_features_stats_wide_df.loc[
            :, 
            ~ total_features_stats_wide_df.columns.get_level_values(level = 'stat').isin(['psi', 'iv'])
            ]
        combined_feature_stats_wide_df = pd.concat([
            self.get_coverage_wide_df(),
            self.get_gini_impurity_wide_df(),
            self.get_auc_wide_df(),
            self.get_corrected_auc_wide_df(),
            self.get_ks_wide_df(),
            self.get_psi_wide_df(),
            self.get_first_psi_wide_df(),
            self.get_adjacent_psi_wide_df(),
            self.get_iv_wide_df(),
            total_features_stats_wide_df, # self.get_total_features_stats_wide_df(),
            self.get_interval_trend_wide_df(),
            self.get_abs_interval_trend_wide_df(),
        ], axis=1)
        if dropna:
            combined_feature_stats_wide_df = combined_feature_stats_wide_df.dropna(how='all', axis=1)
            log.info("Dropped columns with all NaN values.", func = func_name)
        if not (combined_feature_stats_wide_df.columns.is_unique and combined_feature_stats_wide_df.index.is_unique):
            raise ValueError(f'index or columns is duplicated.')
        log.info("Feature stats combined into a wide DataFrame.", func = func_name)

        if feature_importance_df is not None:
            feature_importance_df = feature_importance_df.copy()
            combined_feature_stats_wide_df = combined_feature_stats_wide_df.loc[feature_importance_df.index, :]
            feature_importance_df.columns = get_expanded_header(
                feature_importance_df.columns, 
                'feature_importance'
                )
            combined_feature_stats_wide_df = pd.concat([
                feature_importance_df, 
                combined_feature_stats_wide_df
                ], axis = 1)

        return combined_feature_stats_wide_df
    

    def get_combined_feature_stats_long_df(self, dropna = False):
        func_name = 'get_combined_feature_stats_long_df()'
        log.info("Combining feature stats into a long DataFrame.", func = func_name)
        combined_feature_stats_long_df = pd.concat([
            self.get_distri_long_df(),
            self.get_features_stats_long_df(), 
        ], axis=1)
        if dropna:
            combined_feature_stats_long_df = combined_feature_stats_long_df.dropna(how='all', axis=1)
            log.info("Dropped columns with all NaN values.", func = func_name)
        
        if not (combined_feature_stats_long_df.columns.is_unique and combined_feature_stats_long_df.index.is_unique):
            raise ValueError(f'index or columns is duplicated.')
        log.info("Feature stats combined into a long DataFrame.", func = func_name)
        return combined_feature_stats_long_df
    


    pass



class MonthlyFeatureSelector(MonthlyFeatureStats):

    def __init__(self, 
                 path: str = None,
                 features: str = None,
                 ):
        func_name = 'MonthlyFeatureSelector.init()'
        self.path = path
        self.stat = pickle_load(path)
        self.features = features
        return None
    
    def get_wide_stat_df(self, *args, features: list = None, **kwargs):
        wide_stat_df = self.get_combined_feature_stats_wide_df(*args, **kwargs)
        if self.features is not None:
            wide_stat_df = wide_stat_df.loc[wide_stat_df.index.isin(self.features), :].copy()
        if features is not None:
            wide_stat_df = wide_stat_df.loc[wide_stat_df.index.isin(features), :].copy()
        return wide_stat_df
    
    def get_long_stat_df(self, *args, **kwargs):
        return self.get_combined_feature_stats_long_df(*args, **kwargs)
    
    def describe_wide_stat_df(self, *args, by: str, **kwargs):
        wide_stat_df = self.get_wide_stat_df(*args, **kwargs)
        wide_stat_df = wide_stat_df[by]
        if 'corr_dir' in wide_stat_df.columns:
            log.info(
                f'\ndescribe by: "{by}"', 
                wide_stat_df['corr_dir'].value_counts().reindex([
                    'same', 'opposite', 'maybe_same', 'maybe_opposite', 'unknown'
                    ], fill_value = 0),
                color = 'green')
        wide_stat_df = wide_stat_df[['min', 'max', 'abs_gap', 'rel_gap', 'trend']]
        describe_df = wide_stat_df.describe().T
        stat_cols = ['count', 'min', 'mean', 'max', '25%', '50%', '75%', 'std']
        describe_df = describe_df[stat_cols]
        describe_df['count'] = describe_df['count'].astype(int)
        log.info(f'\ndescribe by: "{by}"', describe_df.round(3), color = 'green')
        return describe_df







###################################################################

# def get_quantile_bins_func(series: pd.Series):
#     return get_quantile_bins(series, drop_endpoints=True, add_endpoints='inf')



# def get_grouped_coverage_long_df(data_df: pd.DataFrame, groups: list, feature_names: list) -> pd.DataFrame:
#     result = (data_df.groupby(by=groups)[feature_names]
#             .agg(lambda x: x.notna().mean())
#             .unstack()
#             .to_frame('coverage'))
#     return result


# def get_grouped_gini_impurity_long_df(
#         data_df: pd.DataFrame, 
#         groups: list,  
#         feature_names: list
#         ):
#     result = (data_df.groupby(by=groups)[feature_names]
#             .agg(calculate_gini_impurity)
#             .unstack()
#             .sort_index()
#             .to_frame('gini_impurity'))
#     return result

# def convert_long_to_monthly_wide_df(
#     data_df: pd.DataFrame,
#     feature_name: str,
#     month: str
# ) -> pd.DataFrame:
#     """
#     Compute monthly statistics and gaps for a single-column DataFrame with MultiIndex.

#     Parameters:
#     - data_df (pd.DataFrame): Input DataFrame with one column and MultiIndex (feature_name and month levels).
#     - feature_name (str): Name of the index level to group by.
#     - month (str): Name of the index level to unstack (typically month).

#     Returns:
#     - pd.DataFrame: DataFrame with monthly values, min, mean, max, absolute gap, and relative gap.

#     Raises:
#     - ValueError: If data_df has more than one column or required index levels are missing.
#     - TypeError: If data_df is not a pandas DataFrame.
#     """
#     # input validation
#     if not isinstance(data_df, pd.DataFrame):
#         raise TypeError("data_df must be a pandas DataFrame")
#     if len(data_df.columns) != 1:
#         raise ValueError("data_df must have exactly one column")
#     if not isinstance(data_df.index, pd.MultiIndex) or feature_name not in data_df.index.names or month not in data_df.index.names:
#         raise ValueError(f"data_df index must be MultiIndex with levels '{feature_name}' and '{month}'")

#     # get data column
#     data_col = data_df.columns[0]

#     # calculate monthly statistics
#     gp = data_df.groupby(level=feature_name)
#     stats_df = pd.concat([
#         gp.min().add_prefix('min-'),
#         gp.mean().add_prefix('mean-'),
#         gp.max().add_prefix('max-'),
#         gp.agg(get_trend_correlation).add_prefix('trend-'),
#     ], axis=1)
    
#     # calculate absolute and relative gaps
#     stats_df[f'abs_gap-{data_col}'] = stats_df[f'max-{data_col}'] - stats_df[f'min-{data_col}']
#     stats_df[f'rel_gap-{data_col}'] = stats_df[f'abs_gap-{data_col}'] / stats_df[f'mean-{data_col}']

#     # unstack monthly values
#     monthly_df = data_df.unstack(level=month)
#     monthly_df.columns = [f'{month_val}-{data_col}' for _, month_val in monthly_df.columns]

#     # concatenate monthly stats and gaps
#     result = pd.concat([monthly_df, stats_df], axis=1)
    
#     return result




# def get_monthly_psi_from_percent_df(percent_df: pd.DataFrame):
#     psi_df = {}
#     months = sorted(percent_df.columns.to_list())
#     if len(months) > 1:
#         for a_month in months:
#             for b_month in months:
#                 if a_month >= b_month:
#                     continue
#                 tmp_df = get_psi_series(
#                     percent_df[a_month],
#                     percent_df[b_month],
#                 )
#                 tmp_df.loc['total'] = tmp_df.sum()
#                 psi_df[f'{a_month}-vs-{b_month}'] = tmp_df
#         psi_df = pd.concat(psi_df)
#         psi_df.index.names = ['months', 'interval']
#     else:
#         log.error('len(months) == 1')
#         raise ValueError('len(months) == 1')
#     return psi_df


# def get_psi_long_df(
#         monthly_distri_stats_df: pd.DataFrame, 
#         disable_multiprocessing = False
#         ):
#     feature_names = monthly_distri_stats_df.index.get_level_values(level = 'feature_name').unique()
#     vars = {
#         col: monthly_distri_stats_df.xs(level = 'feature_name', key = col).xs(level = 'stat', key = 'percent', axis = 1)
#         for col in feature_names
#     }
#     psi_df = pd.concat(multiprocessing_apply(
#         get_monthly_psi_from_percent_df, 
#         vars, 
#         disable_multiprocessing = disable_multiprocessing
#         ))
#     print()
#     if len(psi_df.index.names) == 3:
#         psi_df.index.names = ['feature_name', 'months', 'interval']
#     else:
#         log.error('len(months) == 1 ?')
#         raise ValueError('len(months) == 1 ?')
#     return psi_df


# def drop_level_and_get_iv(data_df: pd.DataFrame, drop_level: str = 'month'):
#     data_df = data_df.droplevel(level = drop_level)
#     iv_df = get_psi(data_df[0], data_df[1], total = True)
#     iv_df.index.name = 'interval'
#     return iv_df

# def get_feature_name_iv_df(stat_df: pd.DataFrame):
#     gp = stat_df.groupby(level='month')
#     iv_df = gp.apply(drop_level_and_get_iv)
#     return iv_df


# def get_iv_long_df(
#         y_distri_df: pd.DataFrame, 
#         feature_names: list, 
#         disable_multiprocessing = False
#         ):
#     feature_names = y_distri_df.index.get_level_values(level = 'feature_name').unique()
#     y_distri_df = y_distri_df.xs(level = 'stat', key = 'percent', axis = 1)
#     vars = {col: y_distri_df.xs(level = 'feature_name', key = col) for col in feature_names}
#     iv_df = multiprocessing_apply(
#         get_feature_name_iv_df, 
#         vars,
#         disable_multiprocessing=disable_multiprocessing
#         )
#     print()
#     iv_df = pd.concat(iv_df)
#     iv_df.index.names = ['feature_name', 'month', 'interval']
#     iv_df.rename(columns = {'psi': 'iv'}, inplace = True)
#     return iv_df


# # def get_iv_stats_df(iv_df: pd.DataFrame):
# #     iv_stats_df = iv_df.xs(level='interval', key = 'total')[['psi']]
# #     iv_stats_df.rename(columns = {'psi': 'iv'}, inplace = True)
# #     iv_stats_df = get_monthly_stats_df(iv_stats_df, 'feature_name', 'month')
# #     return iv_stats_df


# def get_iv_pos_pct_levels(df: pd.DataFrame):
#     df = df.copy()
#     df[('perf', 'pos_pct')] = df[('count', 1)] / df[[('count', 0), ('count', 1)]].sum(axis = 1)
#     df[('perf', 'lift')] = float('nan')
#     df[('perf', 'iv')] = get_psi_series(df[('count', 0)], df[('count', 1)], return_type='series')
#     df.loc['total'] = df.sum()
#     df[('perf', 'lift')] = df[('perf', 'pos_pct')] / df[('perf', 'pos_pct')]['total']
#     df.loc['total', ('perf', 'pos_pct')] = df.loc['total', ('count', 1)] / (df.loc['total', ('count', 0)] + df.loc['total', ('count', 1)])
#     df[[('count', 0), ('count', 1)]] = df[[('count', 0), ('count', 1)]].fillna(0).astype(int)
#     return df


# def get_iv_distri_pos_pct_wide_df(
#         iv_distri_wide_df: pd.DataFrame, 
#         disable_multiprocessing = False
#         ):
#     vars = {
#         (feature_name, month): iv_distri_wide_df
#             .xs(level='feature_name', key = feature_name)
#             .xs(level='month', key = month)
#             for feature_name in iv_distri_wide_df.index.get_level_values(level='feature_name').unique()
#             for month in iv_distri_wide_df.index.get_level_values(level='month').unique()
#             }
#     iv_distri_pos_pct_wide_df = multiprocessing_apply(
#         get_iv_pos_pct_levels, 
#         vars,
#         disable_multiprocessing=disable_multiprocessing
#         )
#     print()
#     iv_distri_pos_pct_wide_df = pd.concat(iv_distri_pos_pct_wide_df)
#     iv_distri_pos_pct_wide_df.index.names = ['feature_name', 'month', 'interval']
#     return iv_distri_pos_pct_wide_df


# #%%

# def get_coverage_dfs(data_df, feature_names):
#     log.info('started', func = 'get_coverage_dfs')
#     coverage_dfs = {}
#     coverage_long_df = get_grouped_coverage_long_df(data_df, 'month', feature_names)
#     coverage_wide_df = convert_long_to_monthly_wide_df(coverage_long_df, 'feature_name', 'month')
#     coverage_dfs['coverage_long_df'] = coverage_long_df
#     coverage_dfs['coverage_wide_df'] = coverage_wide_df
#     log.info('completed', func = 'get_coverage_dfs')
#     return coverage_dfs


# def get_gini_impurity_dfs(data_df, feature_names):
#     log.info('started', func = 'get_gini_impurity_dfs')
#     gini_impurity_dfs = {}
#     gini_impurity_long_df = get_grouped_gini_impurity_long_df(data_df, 'month', feature_names)
#     gini_impurity_wide_df = convert_long_to_monthly_wide_df(gini_impurity_long_df, 'feature_name', 'month')
#     gini_impurity_dfs['gini_impurity_long_df'] = gini_impurity_long_df
#     gini_impurity_dfs['gini_impurity_wide_df'] = gini_impurity_wide_df
#     log.info('completed', func = 'get_gini_impurity_dfs')
#     return gini_impurity_dfs


# def fix_psi_distri_vs_wide_df(df: pd.DataFrame):
#     df = df.droplevel(level = 'feature_name')
#     df.loc['total'] = df.sum()
#     cols = df.columns.get_level_values('stat') == 'count'
#     df.loc[:, cols] = df.loc[:, cols].fillna(0).astype(int)
#     return df

# def get_psi_distri_vs_wide_df(
#         psi_distri_wide_df: pd.DataFrame, 
#         psi_long_df: pd.DataFrame,
#         ):
#     psi_long_df = psi_long_df.copy()
#     psi_distri_wide_df = psi_distri_wide_df.copy()

#     idx = psi_long_df[['psi']].index.get_level_values(level = 'interval') != 'total'
#     psi_months_wide_df = psi_long_df.loc[idx, ['psi']].unstack('months')
#     psi_months_wide_df

#     psi_distri_vs_wide_df = pd.concat([psi_distri_wide_df, psi_months_wide_df], axis = 1)
#     psi_distri_vs_wide_df.columns.names = ['stat', 'month']

#     gp = psi_distri_vs_wide_df.groupby(level = 'feature_name')
#     psi_distri_vs_wide_df = gp.apply(lambda df: fix_psi_distri_vs_wide_df(df))
#     psi_distri_vs_wide_df.columns.names = ['stat', 'month']
#     psi_distri_vs_wide_df

#     cols = psi_distri_vs_wide_df.columns.get_level_values('stat') == 'count'
#     cols = psi_distri_vs_wide_df.columns[cols]
#     for col in cols:
#         psi_distri_vs_wide_df[col] = psi_distri_vs_wide_df[col].fillna(0).astype(int)

#     return psi_distri_vs_wide_df


# def get_psi_dfs(
#         data_df: pd.DataFrame, 
#         feature_names: list, 
#         quantiles_dict: pd.Series, 
#         disable_multiprocessing=False):

#     func_name = 'get_psi_dfs()'
#     log.info('started', func = func_name)
#     psi_dfs = {}

#     if data_df['month'].nunique(dropna=True) < 2:
#         log.error('len(months) < 2')
#         raise ValueError(f'len(months) < 2')

#     group_cols = ['month']
#     log.info(f'get_grouped_distri_long_df()', func = func_name)
#     psi_distri_long_df = get_grouped_distri_long_df(
#         data_df=data_df, 
#         group_cols=group_cols, 
#         feature_names=feature_names, 
#         quantiles_dict=quantiles_dict, 
#         disable_multiprocessing=disable_multiprocessing
#     )
#     psi_dfs['psi_distri_long_df'] = psi_distri_long_df

#     psi_distri_wide_df = psi_distri_long_df.unstack(level = 'month')
#     psi_distri_wide_df.columns.names = ['stat', 'month']
#     psi_dfs['psi_distri_wide_df'] = psi_distri_wide_df

#     log.info('get_psi_long_df()', func = func_name)
#     psi_long_df = get_psi_long_df(
#         psi_distri_wide_df, 
#         disable_multiprocessing = disable_multiprocessing
#         )
#     psi_dfs['psi_long_df'] = psi_long_df

#     psi_distri_vs_wide_df = get_psi_distri_vs_wide_df(psi_distri_wide_df, psi_long_df)
#     psi_dfs['psi_distri_vs_wide_df'] = psi_distri_vs_wide_df

#     psi_long_total_df = psi_long_df.xs(level = 'interval', key = 'total')[['psi']]
#     psi_wide_total_df = convert_long_to_monthly_wide_df(psi_long_total_df, 'feature_name', 'months')
#     psi_wide_total_df['trend-psi'] = float('nan')
#     psi_dfs['psi_wide_total_df'] = psi_wide_total_df

#     months = sorted(set(psi_long_df.index.get_level_values('months').unique().str.split('-vs-').to_series().sum()))
#     first_month = months[0]
#     adjacent_months = [f'{m0}-vs-{m1}' for m0, m1 in zip(months[:-1], months[1:])]

#     idx = psi_long_total_df.index.get_level_values('months').isin(adjacent_months)
#     adjacent_psi_long_total_df = psi_long_total_df.loc[idx, :].copy().add_prefix('adjacent_')
#     psi_dfs['adjacent_psi_long_total_df'] = adjacent_psi_long_total_df

#     adjacent_psi_wide_total_df = convert_long_to_monthly_wide_df(adjacent_psi_long_total_df, 'feature_name', 'months')
#     adjacent_psi_wide_total_df['trend-adjacent_psi'] = float('nan')
#     psi_dfs['adjacent_psi_wide_total_df'] = adjacent_psi_wide_total_df

#     idx = psi_long_total_df.index.get_level_values('months').str.startswith(first_month)
#     first_month_psi_long_total_df = psi_long_total_df.loc[idx, :].copy().add_prefix('first_')
#     psi_dfs['first_month_psi_long_total_df'] = first_month_psi_long_total_df

#     first_psi_wide_total_df = convert_long_to_monthly_wide_df(first_month_psi_long_total_df, 'feature_name', 'months')
#     first_psi_wide_total_df['trend-first_psi'] = float('nan')
#     psi_dfs['first_psi_wide_total_df'] = first_psi_wide_total_df

#     log.info('completed', func = func_name)
#     return psi_dfs


# def get_iv_dfs(data_df, feature_names, quantiles_dict, disable_multiprocessing = False):

#     func_name = 'get_iv_dfs()'
#     log.info('started', func = func_name)
#     iv_dfs = {}

#     group_cols = ['month', 'y']
#     log.info('get_grouped_distri_long_df()', func = func_name)
#     iv_distri_long_df = get_grouped_distri_long_df(
#         data_df=data_df, 
#         group_cols=group_cols, 
#         feature_names=feature_names, 
#         quantiles_dict=quantiles_dict,
#         disable_multiprocessing=disable_multiprocessing
#         )
#     iv_dfs['iv_distri_long_df'] = iv_distri_long_df

#     iv_distri_wide_df = iv_distri_long_df.unstack(level = 'y')
#     iv_distri_wide_df.columns.names = ['stat', 'y']
    
#     log.info('get_iv_distri_pos_pct_wide_df()', func = func_name)
#     iv_distri_pos_pct_wide_df = get_iv_distri_pos_pct_wide_df(
#         iv_distri_wide_df, 
#         disable_multiprocessing = disable_multiprocessing
#         )
#     iv_dfs['iv_distri_pos_pct_wide_df'] = iv_distri_pos_pct_wide_df

#     iv_long_df = iv_distri_pos_pct_wide_df
#     iv_dfs['iv_long_df'] = iv_long_df

#     iv_long_total_df = iv_long_df.xs(level='interval', key = 'total')[[('perf', 'iv')]].droplevel(level = 'stat', axis = 1)
#     iv_wide_total_df = convert_long_to_monthly_wide_df(iv_long_total_df, 'feature_name', 'month')
#     iv_dfs['iv_wide_total_df'] = iv_wide_total_df

#     iv_distri_pos_pct_wide_unstack_df = iv_dfs['iv_distri_pos_pct_wide_df'].unstack(level = 'month')
#     iv_distri_pos_pct_wide_unstack_df.columns = pd.MultiIndex.from_tuples([
#         (f'{stat}-{y}', month) 
#         for stat, y, month 
#         in iv_distri_pos_pct_wide_unstack_df.columns
#         ], names = ['stat', 'month'])
#     iv_dfs['iv_distri_pos_pct_wide_unstack_df'] = iv_distri_pos_pct_wide_unstack_df

#     log.info('completed', func = func_name)

#     return iv_dfs





# class MonthlyFeatureStats(object):

#     def __init__(self, data_df, feature_names, disable_multiprocessing = False, path = None):
#         func_name = 'MonthlyFeatureStats.init()'
#         check_feature_data_df(data_df)
#         self.data_df = data_df
#         self.feature_names = feature_names
#         self.path = path
#         if path is not None and os.path.isfile(path=path):
#             log.info('path isfile', func = func_name)
#             self.stat = pickle_load(path)
#         else:
#             self.stat = {}
#         self.disable_multiprocessing = disable_multiprocessing
#         self.n_months = data_df['month'].nunique(dropna=True)

#         if self.n_months < 2:
#             log.warning('n_months < 2', func = func_name)

#         return None
    

#     def dump(self, makedirs = False):
#         path = self.path
#         stat = self.stat
#         if makedirs:
#             os.makedirs(os.path.dirname(path), exist_ok=True)
#         pickle_dump(stat, path)

#     def get_quantiles_dict(self, renew = False, *args, **kwargs):
#         if 'quantiles_dict' not in self.stat or renew:
#             data_df = self.data_df
#             feature_names = self.feature_names
#             disable_multiprocessing = self.disable_multiprocessing
#             quantiles_dict = get_quantiles_dict(
#                 data_df, feature_names, disable_multiprocessing=disable_multiprocessing
#                 )
#             self.stat['quantiles_dict'] = quantiles_dict
#         else:
#             quantiles_dict = self.stat['quantiles_dict']
#         return quantiles_dict
    
#     def get_coverage_dfs(self, renew = False, *args, **kwargs):
#         if 'coverage_dfs' not in self.stat or renew:
#             data_df = self.data_df
#             feature_names = self.feature_names
#             coverage_dfs = get_coverage_dfs(data_df, feature_names)
#             self.stat['coverage_dfs'] = coverage_dfs
#         else:
#             coverage_dfs = self.stat['coverage_dfs']
#         return coverage_dfs

#     def get_gini_impurity_dfs(self, renew = False, *args, **kwargs):
#         if 'gini_impurity_dfs' not in self.stat or renew:
#             data_df = self.data_df
#             feature_names = self.feature_names
#             gini_impurity_dfs = get_gini_impurity_dfs(data_df, feature_names)
#             self.stat['gini_impurity_dfs'] = gini_impurity_dfs
#         else:
#             gini_impurity_dfs = self.stat['gini_impurity_dfs']
#         return gini_impurity_dfs

#     def get_psi_dfs(self, renew = False, *args, **kwargs):

#         if self.n_months < 2:
#             psi_dfs = {k: pd.DataFrame() for k in [
#                 'psi_distri_long_df',
#                 'psi_distri_wide_df',
#                 'psi_long_df',
#                 'psi_distri_vs_wide_df',
#                 'psi_wide_total_df',
#                 'adjacent_psi_long_total_df',
#                 'adjacent_psi_wide_total_df',
#                 'first_month_psi_long_total_df',
#                 'first_psi_wide_total_df'
#                 ]}
#             self.stat['psi_dfs'] = psi_dfs
#             return psi_dfs
        
#         if 'psi_dfs' not in self.stat or renew:
#             data_df = self.data_df
#             feature_names = self.feature_names
#             quantiles_dict = self.get_quantiles_dict()
#             disable_multiprocessing = self.disable_multiprocessing
#             psi_dfs = get_psi_dfs(
#                 data_df, feature_names, quantiles_dict, disable_multiprocessing=disable_multiprocessing
#                 )
#             self.stat['psi_dfs'] = psi_dfs
#         else:
#             psi_dfs = self.stat['psi_dfs']

#         return psi_dfs

#     def get_iv_dfs(self, renew = False, *args, **kwargs):
#         if 'iv_dfs' not in self.stat or renew:
#             data_df = self.data_df
#             feature_names = self.feature_names
#             quantiles_dict = self.get_quantiles_dict()
#             disable_multiprocessing = self.disable_multiprocessing
#             iv_dfs = get_iv_dfs(
#                 data_df, feature_names, quantiles_dict, 
#                 disable_multiprocessing = disable_multiprocessing)
#             self.stat['iv_dfs'] = iv_dfs
#         else:
#             iv_dfs = self.stat['iv_dfs']
#         return iv_dfs
        
#     def get_simple_wide_df(self, renew = False):
#         func_name = 'get_simple_wide_df()'
#         log.info('started', func = func_name)
#         if 'simple_wide_df' not in self.stat or renew:
#             coverage_dfs = self.get_coverage_dfs()
#             gini_impurity_dfs = self.get_gini_impurity_dfs()
#             psi_dfs = self.get_psi_dfs()
#             iv_dfs = self.get_iv_dfs()
#             simple_wide_dfs = [
#                 coverage_dfs['coverage_wide_df'],
#                 gini_impurity_dfs['gini_impurity_wide_df'],
#                 psi_dfs['psi_wide_total_df'],
#                 psi_dfs['adjacent_psi_wide_total_df'],
#                 psi_dfs['first_psi_wide_total_df'],
#                 iv_dfs['iv_wide_total_df'],
#             ]
#             simple_wide_df = pd.concat(simple_wide_dfs, axis = 1)
#             self.stat['simple_wide_df'] = simple_wide_df
#         else:
#             simple_wide_df = self.stat['simple_wide_df']
#         log.info('completed', func = func_name)
#         return simple_wide_df

#     def get_detail_wide_df(self, renew = False):
#         func_name = 'get_detail_wide_df()'
#         log.info('started', func = func_name)
#         if 'get_detail_wide_df' not in self.stat or renew:
#             psi_dfs = self.get_psi_dfs()
#             iv_dfs = self.get_iv_dfs()
#             psi_distri_vs_wide_df = psi_dfs['psi_distri_vs_wide_df']
#             iv_distri_pos_pct_wide_unstack_df = iv_dfs['iv_distri_pos_pct_wide_unstack_df']
#             if psi_distri_vs_wide_df.empty or psi_distri_vs_wide_df.index.equals(iv_distri_pos_pct_wide_unstack_df.index):
#                 detail_wide_df = pd.concat([psi_distri_vs_wide_df, iv_distri_pos_pct_wide_unstack_df], axis = 1)

#                 uniques = detail_wide_df.columns.get_level_values('month').unique()
#                 if len(uniques) == 1:
#                     month = uniques[0]
#                     detail_wide_df.insert(0, ('count', month), detail_wide_df[('count-0', month)] + detail_wide_df[('count-1', month)])
#                     detail_wide_df.insert(1, ('percent', month), detail_wide_df[('count', month)] / detail_wide_df[('count', month)].xs(level = 'interval', key = 'total'))
#             else:
#                 raise ValueError(f'index.equals = False')
#             self.stat['detail_wide_df'] = detail_wide_df
#         else:
#             detail_wide_df = self.stat['detail_wide_df']
#         log.info('completed', func = func_name)
#         return detail_wide_df
    
#     pass # end of class

